# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for the HTCondor WMS service class and related functions."""

import logging
import os
import unittest
from pathlib import Path
from shutil import copy2

import htcondor
from lsst.ctrl.bps import BpsConfig, GenericWorkflowExec, GenericWorkflowJob, WmsStates
from lsst.ctrl.bps.htcondor.htcondor_config import HTC_DEFAULTS_URI
from lsst.ctrl.bps.htcondor.htcondor_service import (
    HTCondorService,
    JobStatus,
    NodeStatus,
    WmsIdType,
    _get_exit_code_summary,
    _get_info_from_path,
    _get_state_counts_from_dag_job,
    _htc_node_status_to_wms_state,
    _htc_status_to_wms_state,
    _translate_job_cmds,
    _wms_id_to_dir,
)
from lsst.ctrl.bps.htcondor.lssthtc import MISSING_ID
from lsst.utils.tests import temporaryDirectory

logger = logging.getLogger("lsst.ctrl.bps.htcondor")

TESTDIR = os.path.abspath(os.path.dirname(__file__))

LOCATE_SUCCESS = """[
        CondorPlatform = "$CondorPlatform: X86_64-CentOS_7.9 $";
        MyType = "Scheduler";
        Machine = "testmachine";
        Name = "testmachine";
        CondorVersion = "$CondorVersion: 23.0.3 2024-04-04 $";
        MyAddress = "<127.0.0.1:9618?addrs=127.0.0.1-9618+snip>"
    ]
"""

PING_SUCCESS = """[
        AuthCommand = 60011;
        AuthMethods = "FS_REMOTE";
        Command = 60040;
        AuthorizationSucceeded = true;
        ValidCommands = "60002,60003,60011,60014,60045,60046,60047,60048,60049,60050,60052,523";
        TriedAuthentication = true;
        RemoteVersion = "$CondorVersion: 10.9.0 2023-09-28 BuildID: 678228 PackageID: 10.9.0-1 $";
        MyRemoteUserName = "testuser@testmachine";
        Authentication = "YES";
    ]
"""


class HTCondorServiceTestCase(unittest.TestCase):
    """Test selected methods of the HTCondor WMS service class."""

    def setUp(self):
        config = BpsConfig({}, wms_service_class_fqn="lsst.ctrl.bps.htcondor.HTCondorService")
        self.service = HTCondorService(config)

    def tearDown(self):
        pass

    def testDefaults(self):
        self.assertEqual(self.service.defaults["memoryLimit"], 491520)

    def testDefaultsPath(self):
        self.assertEqual(self.service.defaults_uri, HTC_DEFAULTS_URI)
        self.assertFalse(self.service.defaults_uri.isdir())

    @unittest.mock.patch.object(htcondor.Collector, "locate", return_value=LOCATE_SUCCESS)
    @unittest.mock.patch.object(htcondor.SecMan, "ping", return_value=PING_SUCCESS)
    def testPingSuccess(self, mock_locate, mock_ping):
        status, message = self.service.ping(None)
        self.assertEqual(status, 0)
        self.assertEqual(message, "")

    def testPingFailure(self):
        with unittest.mock.patch("htcondor.Collector.locate") as locate_mock:
            locate_mock.side_effect = htcondor.HTCondorLocateError()
            status, message = self.service.ping(None)
            self.assertEqual(status, 1)
            self.assertEqual(message, "Could not locate Schedd service.")

    @unittest.mock.patch.object(htcondor.Collector, "locate", return_value=LOCATE_SUCCESS)
    def testPingPermission(self, mock_locate):
        with unittest.mock.patch("htcondor.SecMan.ping") as ping_mock:
            ping_mock.side_effect = htcondor.HTCondorIOError()
            status, message = self.service.ping(None)
            self.assertEqual(status, 1)
            self.assertEqual(message, "Permission problem with Schedd service.")


class GetExitCodeSummaryTestCase(unittest.TestCase):
    """Test the function responsible for creating exit code summary."""

    def setUp(self):
        self.jobs = {
            "1.0": {
                "JobStatus": htcondor.JobStatus.IDLE,
                "bps_job_label": "foo",
            },
            "2.0": {
                "JobStatus": htcondor.JobStatus.RUNNING,
                "bps_job_label": "foo",
            },
            "3.0": {
                "JobStatus": htcondor.JobStatus.REMOVED,
                "bps_job_label": "foo",
            },
            "4.0": {
                "ExitCode": 0,
                "ExitBySignal": False,
                "JobStatus": htcondor.JobStatus.COMPLETED,
                "bps_job_label": "bar",
            },
            "5.0": {
                "ExitCode": 1,
                "ExitBySignal": False,
                "JobStatus": htcondor.JobStatus.COMPLETED,
                "bps_job_label": "bar",
            },
            "6.0": {
                "ExitBySignal": True,
                "ExitSignal": 11,
                "JobStatus": htcondor.JobStatus.HELD,
                "bps_job_label": "baz",
            },
            "7.0": {
                "ExitBySignal": False,
                "ExitCode": 42,
                "JobStatus": htcondor.JobStatus.HELD,
                "bps_job_label": "baz",
            },
            "8.0": {
                "JobStatus": htcondor.JobStatus.TRANSFERRING_OUTPUT,
                "bps_job_label": "qux",
            },
            "9.0": {
                "JobStatus": htcondor.JobStatus.SUSPENDED,
                "bps_job_label": "qux",
            },
        }

    def tearDown(self):
        pass

    def testMainScenario(self):
        actual = _get_exit_code_summary(self.jobs)
        expected = {"foo": [], "bar": [1], "baz": [11, 42], "qux": []}
        self.assertEqual(actual, expected)

    def testUnknownStatus(self):
        jobs = {
            "1.0": {
                "JobStatus": -1,
                "bps_job_label": "foo",
            }
        }
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            _get_exit_code_summary(jobs)
        self.assertIn("lsst.ctrl.bps.htcondor", cm.records[0].name)
        self.assertIn("Unknown", cm.output[0])
        self.assertIn("JobStatus", cm.output[0])

    def testUnknownKey(self):
        jobs = {
            "1.0": {
                "JobStatus": htcondor.JobStatus.COMPLETED,
                "UnknownKey": None,
                "bps_job_label": "foo",
            }
        }
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            _get_exit_code_summary(jobs)
        self.assertIn("lsst.ctrl.bps.htcondor", cm.records[0].name)
        self.assertIn("Attribute", cm.output[0])
        self.assertIn("not found", cm.output[0])


class HtcNodeStatusToWmsStateTestCase(unittest.TestCase):
    """Test assigning WMS state base on HTCondor node status."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testNotReady(self):
        job = {"NodeStatus": NodeStatus.NOT_READY}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.UNREADY)

    def testReady(self):
        job = {"NodeStatus": NodeStatus.READY}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.READY)

    def testPrerun(self):
        job = {"NodeStatus": NodeStatus.PRERUN}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.MISFIT)

    def testSubmittedHeld(self):
        job = {
            "NodeStatus": NodeStatus.SUBMITTED,
            "JobProcsHeld": 1,
            "StatusDetails": "",
            "JobProcsQueued": 0,
        }
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.HELD)

    def testSubmittedRunning(self):
        job = {
            "NodeStatus": NodeStatus.SUBMITTED,
            "JobProcsHeld": 0,
            "StatusDetails": "not_idle",
            "JobProcsQueued": 0,
        }
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.RUNNING)

    def testSubmittedPending(self):
        job = {
            "NodeStatus": NodeStatus.SUBMITTED,
            "JobProcsHeld": 0,
            "StatusDetails": "",
            "JobProcsQueued": 1,
        }
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.PENDING)

    def testPostrun(self):
        job = {"NodeStatus": NodeStatus.POSTRUN}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.MISFIT)

    def testDone(self):
        job = {"NodeStatus": NodeStatus.DONE}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.SUCCEEDED)

    def testErrorDagmanSuccess(self):
        job = {"NodeStatus": NodeStatus.ERROR, "StatusDetails": "DAGMAN error 0"}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.SUCCEEDED)

    def testErrorDagmanFailure(self):
        job = {"NodeStatus": NodeStatus.ERROR, "StatusDetails": "DAGMAN error 1"}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.FAILED)

    def testFutile(self):
        job = {"NodeStatus": NodeStatus.FUTILE}
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.PRUNED)

    def testDeletedJob(self):
        job = {
            "NodeStatus": NodeStatus.ERROR,
            "StatusDetails": "HTCondor reported ULOG_JOB_ABORTED event for job proc (1.0.0)",
            "JobProcsQueued": 0,
        }
        result = _htc_node_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.DELETED)


class HtcStatusToWmsStateTestCase(unittest.TestCase):
    """Test assigning WMS state base on HTCondor status."""

    def testJobStatus(self):
        job = {
            "ClusterId": 1,
            "JobStatus": htcondor.JobStatus.IDLE,
            "bps_job_label": "foo",
        }
        result = _htc_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.PENDING)

    def testNodeStatus(self):
        # Hold/Release test case
        job = {
            "ClusterId": 1,
            "JobStatus": 0,
            "NodeStatus": NodeStatus.SUBMITTED,
            "JobProcsHeld": 0,
            "StatusDetails": "",
            "JobProcsQueued": 1,
        }
        result = _htc_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.PENDING)

    def testNeitherStatus(self):
        job = {"ClusterId": 1}
        result = _htc_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.MISFIT)

    def testRetrySuccess(self):
        job = {
            "NodeStatus": 5,
            "Node": "8e62c569-ae2e-44e8-be36-d1aee333a129_isr_903342_10",
            "RetryCount": 0,
            "ClusterId": 851,
            "ProcId": 0,
            "MyType": "JobTerminatedEvent",
            "EventTypeNumber": 5,
            "HoldReasonCode": 3,
            "HoldReason": "Job raised a signal 9. Handling signal as if job has gone over memory limit.",
            "HoldReasonSubCode": 34,
            "ToE": {
                "ExitBySignal": False,
                "ExitCode": 0,
            },
            "JobStatus": JobStatus.COMPLETED,
            "ExitBySignal": False,
            "ExitCode": 0,
        }
        result = _htc_status_to_wms_state(job)
        self.assertEqual(result, WmsStates.SUCCEEDED)


class TranslateJobCmdsTestCase(unittest.TestCase):
    """Test _translate_job_cmds method."""

    def setUp(self):
        self.gw_exec = GenericWorkflowExec("test_exec", "/dummy/dir/pipetask")
        self.cached_vals = {"profile": {}}

    def testRetryUnlessNone(self):
        gwjob = GenericWorkflowJob("retryUnless", executable=self.gw_exec)
        gwjob.retry_unless_exit = None
        htc_commands = _translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertNotIn("retry_until", htc_commands)

    def testRetryUnlessInt(self):
        gwjob = GenericWorkflowJob("retryUnlessInt", executable=self.gw_exec)
        gwjob.retry_unless_exit = 3
        htc_commands = _translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(int(htc_commands["retry_until"]), gwjob.retry_unless_exit)

    def testRetryUnlessList(self):
        gwjob = GenericWorkflowJob("retryUnlessList", executable=self.gw_exec)
        gwjob.retry_unless_exit = [1, 2]
        htc_commands = _translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertEqual(htc_commands["retry_until"], "member(ExitCode, {1,2})")

    def testRetryUnlessBad(self):
        gwjob = GenericWorkflowJob("retryUnlessBad", executable=self.gw_exec)
        gwjob.retry_unless_exit = "1,2,3"
        with self.assertRaises(ValueError) as cm:
            _ = _translate_job_cmds(self.cached_vals, None, gwjob)
        self.assertIn("retryUnlessExit", str(cm.exception))


class GetStateCountsFromDagJobTestCase(unittest.TestCase):
    """Test counting number of jobs per WMS state."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testCounts(self):
        job = {
            "DAG_NodesDone": 1,
            "DAG_JobsHeld": 2,
            "DAG_NodesFailed": 3,
            "DAG_NodesFutile": 4,
            "DAG_NodesQueued": 5,
            "DAG_NodesReady": 0,
            "DAG_NodesUnready": 7,
            "DAG_NodesTotal": 22,
        }

        truth = {
            WmsStates.SUCCEEDED: 1,
            WmsStates.HELD: 2,
            WmsStates.UNREADY: 7,
            WmsStates.READY: 0,
            WmsStates.FAILED: 3,
            WmsStates.PRUNED: 4,
            WmsStates.MISFIT: 0,
        }

        total, result = _get_state_counts_from_dag_job(job)
        self.assertEqual(total, 22)
        self.assertEqual(result, truth)


class GetInfoFromPathTestCase(unittest.TestCase):
    """Test _get_info_from_path function"""

    def test_tmpdir_abort(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_tmpdir_abort.dag.dagman.out", tmp_dir)
            wms_workflow_id, jobs, message = _get_info_from_path(tmp_dir)
            self.assertEqual(wms_workflow_id, MISSING_ID)
            self.assertEqual(jobs, {})
            self.assertIn("Cannot submit from /tmp", message)

    def test_no_dagman_messages(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_no_messages.dag.dagman.out", tmp_dir)
            wms_workflow_id, jobs, message = _get_info_from_path(tmp_dir)
            self.assertEqual(wms_workflow_id, MISSING_ID)
            self.assertEqual(jobs, {})
            self.assertIn("Could not find HTCondor files", message)

    def test_successful_run(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.log", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.out", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.nodes.log", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.node_status", tmp_dir)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.info.json", tmp_dir)
            wms_workflow_id, jobs, message = _get_info_from_path(tmp_dir)
            self.assertEqual(wms_workflow_id, "1163.0")
            self.assertEqual(len(jobs), 6)  # dag, pipetaskInit, 3 science, finalJob
            self.assertEqual(message, "")

    def test_relative_path(self):
        orig_dir = Path.cwd()
        with temporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            abs_path = Path(tmp_dir).resolve() / "subdir"
            abs_path.mkdir()
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.log", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.dagman.out", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.dag.nodes.log", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.node_status", abs_path)
            copy2(f"{TESTDIR}/data/test_pipelines_check_20240727T003507Z.info.json", abs_path)
            wms_workflow_id, jobs, message = _get_info_from_path("subdir")
            self.assertEqual(wms_workflow_id, "1163.0")
            self.assertEqual(len(jobs), 6)  # dag, pipetaskInit, 3 science, finalJob
            self.assertEqual(message, "")
            self.assertEqual(jobs["1163.0"]["Iwd"], str(abs_path))
            os.chdir(orig_dir)


class WmsIdToDirTestCase(unittest.TestCase):
    """Test _wms_id_to_dir function"""

    @unittest.mock.patch("lsst.ctrl.bps.htcondor.htcondor_service._wms_id_type")
    def testInvalidIdType(self, _wms_id_type_mock):
        _wms_id_type_mock.return_value = WmsIdType.UNKNOWN
        with self.assertRaises(TypeError) as cm:
            _, _ = _wms_id_to_dir("not_used")
        self.assertIn("Invalid job id type", str(cm.exception))

    @unittest.mock.patch("lsst.ctrl.bps.htcondor.htcondor_service._wms_id_type")
    def testAbsPathId(self, mock_wms_id_type):
        mock_wms_id_type.return_value = WmsIdType.PATH
        with temporaryDirectory() as tmp_dir:
            wms_path, id_type = _wms_id_to_dir(tmp_dir)
            self.assertEqual(id_type, WmsIdType.PATH)
            self.assertEqual(Path(tmp_dir).resolve(), wms_path)

    @unittest.mock.patch("lsst.ctrl.bps.htcondor.htcondor_service._wms_id_type")
    def testRelPathId(self, _wms_id_type_mock):
        _wms_id_type_mock.return_value = WmsIdType.PATH
        orig_dir = Path.cwd()
        with temporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            abs_path = Path(tmp_dir) / "newdir"
            abs_path.mkdir()
            wms_path, id_type = _wms_id_to_dir("newdir")
            self.assertEqual(id_type, WmsIdType.PATH)
            self.assertEqual(abs_path.resolve(), wms_path)
            os.chdir(orig_dir)



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Unit tests for classes and functions in lssthtc.py"""

import logging
import os
import pathlib
import tempfile
import unittest
from shutil import copy2

import htcondor
from lsst.ctrl.bps.htcondor import lssthtc
from lsst.utils.tests import temporaryDirectory

logger = logging.getLogger("lsst.ctrl.bps.htcondor")
TESTDIR = os.path.abspath(os.path.dirname(__file__))


class TestLsstHtc(unittest.TestCase):
    """Test basic usage."""

    def testHtcEscapeInt(self):
        self.assertEqual(lssthtc.htc_escape(100), 100)

    def testHtcEscapeDouble(self):
        self.assertEqual(lssthtc.htc_escape('"double"'), '""double""')

    def testHtcEscapeSingle(self):
        self.assertEqual(lssthtc.htc_escape("'single'"), "''single''")

    def testHtcEscapeNoSideEffect(self):
        val = "'val'"
        self.assertEqual(lssthtc.htc_escape(val), "''val''")
        self.assertEqual(val, "'val'")

    def testHtcEscapeQuot(self):
        self.assertEqual(lssthtc.htc_escape("&quot;val&quot;"), '"val"')

    def testHtcVersion(self):
        ver = lssthtc.htc_version()
        self.assertRegex(ver, r"^\d+\.\d+\.\d+$")


class TweakJobInfoTestCase(unittest.TestCase):
    """Test the function responsible for massaging job information."""

    def setUp(self):
        self.log_file = tempfile.NamedTemporaryFile(prefix="test_", suffix=".log")
        self.log_name = pathlib.Path(self.log_file.name)
        self.job = {
            "Cluster": 1,
            "Proc": 0,
            "Iwd": str(self.log_name.parent),
            "Owner": self.log_name.owner(),
            "MyType": None,
            "TerminatedNormally": True,
        }

    def tearDown(self):
        self.log_file.close()

    def testDirectAssignments(self):
        lssthtc._tweak_log_info(self.log_name, self.job)
        self.assertEqual(self.job["ClusterId"], self.job["Cluster"])
        self.assertEqual(self.job["ProcId"], self.job["Proc"])
        self.assertEqual(self.job["Iwd"], str(self.log_name.parent))
        self.assertEqual(self.job["Owner"], self.log_name.owner())

    def testJobStatusAssignmentJobAbortedEvent(self):
        job = self.job | {"MyType": "JobAbortedEvent"}
        lssthtc._tweak_log_info(self.log_name, job)
        self.assertTrue("JobStatus" in job)
        self.assertEqual(job["JobStatus"], htcondor.JobStatus.REMOVED)

    def testJobStatusAssignmentExecuteEvent(self):
        job = self.job | {"MyType": "ExecuteEvent"}
        lssthtc._tweak_log_info(self.log_name, job)
        self.assertTrue("JobStatus" in job)
        self.assertEqual(job["JobStatus"], htcondor.JobStatus.RUNNING)

    def testJobStatusAssignmentSubmitEvent(self):
        job = self.job | {"MyType": "SubmitEvent"}
        lssthtc._tweak_log_info(self.log_name, job)
        self.assertTrue("JobStatus" in job)
        self.assertEqual(job["JobStatus"], htcondor.JobStatus.IDLE)

    def testJobStatusAssignmentJobHeldEvent(self):
        job = self.job | {"MyType": "JobHeldEvent"}
        lssthtc._tweak_log_info(self.log_name, job)
        self.assertTrue("JobStatus" in job)
        self.assertEqual(job["JobStatus"], htcondor.JobStatus.HELD)

    def testJobStatusAssignmentJobTerminatedEvent(self):
        job = self.job | {"MyType": "JobTerminatedEvent"}
        lssthtc._tweak_log_info(self.log_name, job)
        self.assertTrue("JobStatus" in job)
        self.assertEqual(job["JobStatus"], htcondor.JobStatus.COMPLETED)

    def testJobStatusAssignmentPostScriptTerminatedEvent(self):
        job = self.job | {"MyType": "PostScriptTerminatedEvent"}
        lssthtc._tweak_log_info(self.log_name, job)
        self.assertTrue("JobStatus" in job)
        self.assertEqual(job["JobStatus"], htcondor.JobStatus.COMPLETED)

    def testAddingExitStatusSuccess(self):
        job = self.job | {
            "MyType": "JobTerminatedEvent",
            "ToE": {"ExitBySignal": False, "ExitCode": 1},
        }
        lssthtc._tweak_log_info(self.log_name, job)
        self.assertIn("ExitBySignal", job)
        self.assertIs(job["ExitBySignal"], False)
        self.assertIn("ExitCode", job)
        self.assertEqual(job["ExitCode"], 1)

    def testAddingExitStatusFailure(self):
        job = self.job | {
            "MyType": "JobHeldEvent",
        }
        with self.assertLogs(logger=logger, level="ERROR") as cm:
            lssthtc._tweak_log_info(self.log_name, job)
        self.assertIn("Could not determine exit status", cm.output[0])

    def testLoggingUnknownLogEvent(self):
        job = self.job | {"MyType": "Foo"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            lssthtc._tweak_log_info(self.log_name, job)
        self.assertIn("Unknown log event", cm.output[1])

    def testMissingKey(self):
        job = self.job
        del job["Cluster"]
        with self.assertRaises(KeyError) as cm:
            lssthtc._tweak_log_info(self.log_name, job)
        self.assertEqual(str(cm.exception), "'Cluster'")


class HtcCheckDagmanOutputTestCase(unittest.TestCase):
    """Test htc_check_dagman_output function."""

    def test_missing_output_file(self):
        with temporaryDirectory() as tmp_dir:
            with self.assertRaises(FileNotFoundError):
                _ = lssthtc.htc_check_dagman_output(tmp_dir)

    def test_permissions_output_file(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_tmpdir_abort.dag.dagman.out", tmp_dir)
            os.chmod(f"{tmp_dir}/test_tmpdir_abort.dag.dagman.out", 0o200)
            print(os.stat(f"{tmp_dir}/test_tmpdir_abort.dag.dagman.out"))
            results = lssthtc.htc_check_dagman_output(tmp_dir)
            os.chmod(f"{tmp_dir}/test_tmpdir_abort.dag.dagman.out", 0o600)
            self.assertIn("Could not read dagman output file", results)

    def test_submit_failure(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/bad_submit.dag.dagman.out", tmp_dir)
            results = lssthtc.htc_check_dagman_output(tmp_dir)
            self.assertIn("Warn: Job submission issues (last: ", results)

    def test_tmpdir_abort(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_tmpdir_abort.dag.dagman.out", tmp_dir)
            results = lssthtc.htc_check_dagman_output(tmp_dir)
            self.assertIn("Cannot submit from /tmp", results)

    def test_no_messages(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/test_no_messages.dag.dagman.out", tmp_dir)
            results = lssthtc.htc_check_dagman_output(tmp_dir)
            self.assertEqual("", results)


class SummaryFromDagTestCase(unittest.TestCase):
    """Test summary_from_dag function."""

    def test_no_dag_file(self):
        with temporaryDirectory() as tmp_dir:
            summary, job_name_to_pipetask = lssthtc.summary_from_dag(tmp_dir)
            self.assertFalse(len(job_name_to_pipetask))
            self.assertFalse(summary)

    def test_success(self):
        with temporaryDirectory() as tmp_dir:
            copy2(f"{TESTDIR}/data/good.dag", tmp_dir)
            summary, job_name_to_pipetask = lssthtc.summary_from_dag(tmp_dir)
            self.assertEqual(summary, "pipetaskInit:1;label1:1;label2:1;label3:1;finalJob:1")
            self.assertEqual(
                job_name_to_pipetask,
                {
                    "pipetaskInit": "pipetaskInit",
                    "0682f8f9-12f0-40a5-971e-8b30c7231e5c_label1_val1_val2": "label1",
                    "d0305e2d-f164-4a85-bd24-06afe6c84ed9_label2_val1_val2": "label2",
                    "2806ecc9-1bba-4362-8fff-ab4e6abb9f83_label3_val1_val2": "label3",
                    "finalJob": "finalJob",
                },
            )


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for job ClassAd handlers."""

import logging
import unittest
from typing import Any

from lsst.ctrl.bps.htcondor.handlers import (
    Chain,
    Handler,
    JobCompletedWithExecTicketHandler,
    JobCompletedWithoutExecTicketHandler,
    JobHeldByOtherHandler,
    JobHeldBySignalHandler,
    JobHeldByUserHandler,
)

logger = logging.getLogger("lsst.ctrl.bps.htcondor")


class DummyHandler(Handler):
    """A concrete handler that does nothing."""

    def handle(self, ad: dict[str, Any]) -> dict[str, Any]:
        pass


class RaisingHandler(Handler):
    """A concrete handler that raises KeyError exception."""

    def handle(self, ad: dict[str, Any]) -> dict[str, Any]:
        raise KeyError("foo")


class ChainTestCase(unittest.TestCase):
    """Test the Chain class."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testDefaultInitialization(self):
        chain = Chain()
        self.assertEqual(len(chain), 0)

    def testCustomInitialization(self):
        handler = DummyHandler()
        chain = Chain(handlers=[handler])
        self.assertEqual(len(chain), 1)
        self.assertIs(chain[0], handler)

    def testAppendingHandler(self):
        first = DummyHandler()
        second = DummyHandler()
        chain = Chain(handlers=[first])
        chain.append(second)
        self.assertEqual(len(chain), 2)
        self.assertIs(chain[0], first)
        self.assertIs(chain[1], second)

    def testAppendingNonHandler(self):
        handler = "foo"
        chain = Chain()
        with self.assertRaises(TypeError):
            chain.append(handler)


class JobCompletedWithExecTicketHandlerTestCase(unittest.TestCase):
    """Test the handler for a completed job with the ticket of execution."""

    def setUp(self):
        self.ad = {"ClusterId": 1, "ProcId": 0, "MyType": "JobTerminatedEvent"}
        self.handler = JobCompletedWithExecTicketHandler()

    def tearDown(self):
        pass

    def testNormalTermination(self):
        ad = self.ad | {"ToE": {"ExitBySignal": False, "ExitCode": 0}}
        result = self.handler.handle(ad)
        self.assertIsNotNone(result)
        self.assertIn("ExitBySignal", result)
        self.assertFalse(result["ExitBySignal"])
        self.assertIn("ExitCode", result)
        self.assertEqual(result["ExitCode"], 0)

    def testAbnormalTermination(self):
        ad = self.ad | {"ToE": {"ExitBySignal": True, "ExitSignal": 9}}
        result = self.handler.handle(ad)
        self.assertIsNotNone(result)
        self.assertIn("ExitBySignal", result)
        self.assertTrue(result["ExitBySignal"])
        self.assertIn("ExitSignal", result)
        self.assertEqual(result["ExitSignal"], 9)

    def testNotHandlingMissingExecTicket(self):
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(self.ad)
        self.assertIsNone(result)
        self.assertIn("ticket of execution", cm.output[0])
        self.assertIn("missing", cm.output[0])

    def testNotHandlingJobNotCompleted(self):
        ad = self.ad | {"MyType": "foo"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("job not completed", cm.output[0])


class JobCompletedWithoutExecTicketHandlerTestCase(unittest.TestCase):
    """Test the handler for a completed job w/o the ticket of execution."""

    def setUp(self):
        self.ad = {"ClusterId": 1, "ProcId": 0, "MyType": "JobTerminatedEvent"}
        self.handler = JobCompletedWithoutExecTicketHandler()

    def tearDown(self):
        pass

    def testNormalTermination(self):
        ad = self.ad | {"TerminatedNormally": True, "ReturnValue": 0}
        result = self.handler.handle(ad)
        self.assertIsNotNone(result)
        self.assertIn("ExitBySignal", result)
        self.assertFalse(result["ExitBySignal"])
        self.assertIn("ExitCode", result)
        self.assertEqual(result["ExitCode"], 0)

    def testAbnormalTermination(self):
        ad = self.ad | {"TerminatedNormally": False, "TerminatedBySignal": 9}
        result = self.handler.handle(ad)
        self.assertIsNotNone(result)
        self.assertIn("ExitBySignal", result)
        self.assertTrue(result["ExitBySignal"])
        self.assertIn("ExitSignal", result)
        self.assertEqual(result["ExitSignal"], 9)

    def testNotHandlingExecTicketExists(self):
        ad = self.ad | {"ToE": {"ExitBySignal": False, "ExitCode": 0}}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("ticket of execution", cm.output[0])
        self.assertIn("found", cm.output[0])

    def testNotHandlingJobNotCompleted(self):
        ad = self.ad | {"MyType": "foo"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("job not completed", cm.output[0])


class JobHeldOtherTestCase(unittest.TestCase):
    """Test the handler for a held job."""

    def setUp(self):
        self.ad = {"ClusterId": 1, "ProcId": 0, "MyType": "JobHeldEvent"}
        self.handler = JobHeldByOtherHandler()

    def tearDown(self):
        pass

    def testHeld(self):
        ad = self.ad | {"HoldReasonCode": 42}
        result = self.handler.handle(ad)
        self.assertIsNotNone(result)
        self.assertIn("ExitBySignal", result)
        self.assertFalse(result["ExitBySignal"])
        self.assertIn("ExitCode", result)
        self.assertEqual(result["ExitCode"], 42)

    def testHeldBySignal(self):
        ad = self.ad | {"HoldReasonCode": 3}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("invalid hold reason code", cm.output[0])
        self.assertIn("HoldReasonCode = 3", cm.output[0])

    def testHeldByUser(self):
        ad = self.ad | {"HoldReasonCode": 1}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("invalid hold reason code", cm.output[0])
        self.assertIn("HoldReasonCode = 1", cm.output[0])

    def testNotHandlingJobNotHeld(self):
        ad = self.ad | {"MyType": "foo"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("job not held", cm.output[0])


class JobHeldBySignalHandlerTestCase(unittest.TestCase):
    """Test the handler for a job held by a signal."""

    def setUp(self):
        self.ad = {"ClusterId": 1, "ProcId": 0, "MyType": "JobHeldEvent"}
        self.handler = JobHeldBySignalHandler()

    def tearDown(self):
        pass

    def testSignalAvailable(self):
        ad = self.ad | {"HoldReasonCode": 3, "HoldReason": "Job raised a signal 9."}
        result = self.handler.handle(ad)
        self.assertIsNotNone(ad)
        self.assertIn("ExitBySignal", result)
        self.assertTrue(result["ExitBySignal"])
        self.assertIn("ExitSignal", result)
        self.assertEqual(int(result["ExitSignal"]), 9)

    def testSignalNotAvailable(self):
        ad = self.ad | {"HoldReasonCode": 3, "HoldReason": "foo"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("signal not found", cm.output[0])

    def testNotHandlingInvalidHoldReasonCode(self):
        ad = self.ad | {"HoldReasonCode": 1, "HoldReason": "via condor_hold (by user foo)"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("not held by a signal", cm.output[0])

    def testNotHandlingJobNotHeld(self):
        ad = self.ad | {"MyType": "foo"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("job not held", cm.output[0])


class JobHeldByUserHandlerTestCase(unittest.TestCase):
    """Test the handler for a job held by the user."""

    def setUp(self):
        self.ad = {"ClusterId": 1, "ProcId": 0, "MyType": "JobHeldEvent"}
        self.handler = JobHeldByUserHandler()

    def tearDown(self):
        pass

    def testHandling(self):
        ad = self.ad | {"HoldReasonCode": 1}
        result = self.handler.handle(ad)
        self.assertIsNotNone(result)
        self.assertIn("ExitBySignal", result)
        self.assertFalse(result["ExitBySignal"])
        self.assertIn("ExitCode", result)
        self.assertEqual(result["ExitCode"], 0)

    def testNotHandlingInvalidHoldReaconCode(self):
        ad = self.ad | {"HoldReasonCode": 3, "HoldReason": "Job raised a signal 9."}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("not held by the user", cm.output[0])

    def testNotHandlingJobNotHeld(self):
        ad = self.ad | {"MyType": "foo"}
        with self.assertLogs(logger=logger, level="DEBUG") as cm:
            result = self.handler.handle(ad)
        self.assertIsNone(result)
        self.assertIn("job not held", cm.output[0])



# This file is part of ctrl_bps_htcondor.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Definitions of handlers of HTCondor job ClassAds."""

__all__ = [
    "HTC_JOB_AD_HANDLERS",
    "Chain",
    "Handler",
    "JobCompletedWithExecTicketHandler",
    "JobCompletedWithoutExecTicketHandler",
    "JobHeldByOtherHandler",
    "JobHeldBySignalHandler",
    "JobHeldByUserHandler",
]


import abc
import logging
import re
from collections.abc import Sequence
from typing import Any

_LOG = logging.getLogger(__name__)


class Handler(abc.ABC):
    """Abstract base class defining Handler interface."""

    @abc.abstractmethod
    def handle(self, ad: dict[str, Any]) -> dict[str, Any] | None:
        """Handle a ClassAd.

        Parameters
        ----------
        ad : `dict[`str`, Any]`
            The dictionary representing ClassAd that need to be processed.

        Returns
        -------
        ad : `dict[`str`, Any]` | None
            The dictionary representing ClassAd after processing and ``None``
            if the handler was not able to process the ad.

        Notes
        -----
        To optimize the memory usage, the implementation of this method may
        modify the ClassAd in place. In such a case, the ClassAd returned by
        the method will be the same object that was passed to it as
        the argument, but including any modifications that were made.
        """


class Chain(Sequence):
    """Class defining chaining of handlers.

    Parameters
    ----------
    handlers : `Sequence` [`Handler`]
        List of handlers that will be used to initialize the chain.
    """

    def __init__(self, handlers: Sequence[Handler] = None) -> None:
        self._handlers = []
        if handlers is not None:
            for handler in handlers:
                self.append(handler)

    def __getitem__(self, index: int) -> Handler:
        return self._handlers[index]

    def __len__(self) -> int:
        return len(self._handlers)

    def append(self, handler: Handler) -> None:
        """Append a handler to the chain.

        Parameters
        ----------
        handler : `Handler`
            The handler that needs to be added to the chain.

        Raises
        ------
        TypeError
            Raised if the passed object in not a ``Handler``.
        """
        if not isinstance(handler, Handler):
            raise TypeError(f"append() argument must be a 'Handler', not a '{type(handler)}'")
        self._handlers.append(handler)

    def handle(self, ad: dict[str, Any]) -> dict[str, Any] | None:
        """Handle a ClassAd.

        Parameters
        ----------
        ad : `dict[`str`, Any]`
            The dictionary representing a ClassAd that need to be handled.

        Returns
        -------
        ad : `dict[`str`, Any]`
            A modified ClassAd if any handler in the chain was able to
            process the ad, None otherwise.
        """
        new_ad = None
        for handler in self:
            try:
                new_ad = handler.handle(ad)
            except Exception as e:
                _LOG.warning(
                    "Handler '%s' raised an exception while processing the ad: %s. "
                    "Proceeding to the next handler (if any).",
                    type(handler).__name__,
                    repr(e),
                )
            else:
                if new_ad is not None:
                    break
        return new_ad


class JobCompletedWithExecTicketHandler(Handler):
    """Handler of ClassAds for completed jobs with the ticket of execution.

    Usually, the entry in the event log for a completed job contains the ticket
    of execution -- a record describing how and when the job was terminated.
    If it exists, this handler will use it to add the attributes describing
    job's exit status.
    """

    def handle(self, ad: dict[str, Any]) -> dict[str, Any] | None:
        if not ad["MyType"].endswith("TerminatedEvent"):
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': job not completed",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
            )
            return None
        if "ToE" in ad:
            toe = ad["ToE"]
            ad["ExitBySignal"] = toe["ExitBySignal"]
            if ad["ExitBySignal"]:
                ad["ExitSignal"] = toe["ExitSignal"]
            else:
                ad["ExitCode"] = toe["ExitCode"]
        else:
            _LOG.debug(
                "%s: refusing to process the ad for the job '%s.%s': ticket of execution missing",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
            )
            return None
        return ad


class JobCompletedWithoutExecTicketHandler(Handler):
    """Handler of ClassAds for completed jobs w/o the ticket of execution.

    The entry in the event log for some completed jobs (e.g. jobs that run
    ``condor_dagman``) do *not* contain the ticket of execution -- a record
    describing how and when the job was terminated.  This handler will try
    to use other existing attributes to add the ones describing job's exit
    status.
    """

    def handle(self, ad: dict[str, Any]) -> dict[str, Any] | None:
        if not ad["MyType"].endswith("TerminatedEvent"):
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': job not completed",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
            )
            return None
        if "ToE" not in ad:
            ad["ExitBySignal"] = not ad["TerminatedNormally"]
            if ad["ExitBySignal"]:
                ad["ExitSignal"] = ad["TerminatedBySignal"]
            else:
                ad["ExitCode"] = ad["ReturnValue"]
        else:
            _LOG.debug(
                "%s: refusing to process the ad for the job '%s.%s': ticket of execution found",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
            )
            return None
        return ad


class JobHeldByOtherHandler(Handler):
    """Handler of ClassAds for jobs put on hold."""

    def handle(self, ad: dict[str, Any]) -> dict[str, Any] | None:
        if not ad["MyType"].endswith("HeldEvent"):
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': job not held",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
            )
            return None
        if ad["HoldReasonCode"] not in {1, 3}:
            ad["ExitBySignal"] = False
            ad["ExitCode"] = ad["HoldReasonCode"]
        else:
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': "
                "invalid hold reason code: HoldReasonCode = %s",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
                ad["HoldReasonCode"],
            )
            return None
        return ad


class JobHeldBySignalHandler(Handler):
    """Handler of ClassAds for jobs put on hold by signals."""

    def handle(self, ad: dict[str, Any]) -> dict[str, Any] | None:
        if not ad["MyType"].endswith("HeldEvent"):
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': job not held",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
            )
            return None
        if ad["HoldReasonCode"] == 3:
            match = re.search(r"signal (\d+)", ad["HoldReason"])
            if match is not None:
                ad["ExitBySignal"] = True
                ad["ExitSignal"] = match.group(1)
            else:
                _LOG.debug(
                    "Handler '%s': refusing to process the ad for the job '%s.%s': "
                    "signal not found: HoldReason = %s",
                    self.__class__.__name__,
                    ad["ClusterId"],
                    ad["ProcId"],
                    ad["HoldReason"],
                )
                return None
        else:
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': "
                "job not held by a signal: HoldReasonCode = %s, HoldReason = %s",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
                ad["HoldReasonCode"],
                ad["HoldReason"],
            )
            return None
        return ad


class JobHeldByUserHandler(Handler):
    """Handler of ClassAds for jobs put on hold by the user."""

    def handle(self, ad: dict[str, Any]) -> dict[str, Any] | None:
        if not ad["MyType"].endswith("HeldEvent"):
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': job not held",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
            )
            return None
        if ad["HoldReasonCode"] == 1:
            ad["ExitBySignal"] = False
            ad["ExitCode"] = 0
        else:
            _LOG.debug(
                "Handler '%s': refusing to process the ad for the job '%s.%s': "
                "job not held by the user: HoldReasonCode = %s, HoldReason = %s",
                self.__class__.__name__,
                ad["ClusterId"],
                ad["ProcId"],
                ad["HoldReasonCode"],
                ad["HoldReason"],
            )
            return None
        return ad


_handlers = [
    JobHeldByUserHandler(),
    JobHeldBySignalHandler(),
    JobHeldByOtherHandler(),
    JobCompletedWithExecTicketHandler(),
    JobCompletedWithoutExecTicketHandler(),
]
HTC_JOB_AD_HANDLERS = Chain(handlers=_handlers)
